"""
Copyright (c) 2009 Ableton AG

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""


class OverlapError(Exception):
    pass


class Replacer(object):


    def __init__(self, content):
        self.content = content
        self.replacements = []

    def __setitem__(self, position, replacement):
        assert position.step is None
        lc = len(self.content)
        start, stop = position.start, position.stop
        # adapt position to size of our actual
        # content
        if start is None:
            start = 0
        if stop is None:
            stop = lc
        if start < 0:
            start += lc
        if stop < 0:
            stop += lc

        if not stop <= lc:
            raise IndexError, "Assignemnt only within content size"

        position = slice(start, stop)
        self.replacements.append((position, replacement))


    def get_value(self):
        return "".join(iter(self))


    def __iter__(self):
        content = self.content
        replacements = sorted(self.replacements)
        if not replacements:
            yield content
            raise StopIteration
        # check for overlaps
        until = -1
        for position, _ in replacements:
            if position.start < until:
                raise OverlapError()
            until = position.stop
        offset = 0
        sit = iter(replacements)
        position, chunk = sit.next()


        if position.start > 0:
            yield content[:position.start]
        while True:
            if callable(chunk):
                chunk = chunk()
            yield chunk
            stop = position.stop
            try:
                position, chunk = sit.next()
            except StopIteration:
                break
            if position.start > stop:
                yield content[stop:position.start]
        if stop < len(content):
            yield content[stop:]






"""
Copyright (c) 2009 Ableton AG

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import absolute_import, with_statement

import sys
from functools import wraps
import codecs
import logging
import pprint
from collections import defaultdict


from pyparsing import originalTextFor, ParseException

from .parser import (
    CSSParser,
    StringNode,
    URINode,
    )

from .util import Replacer

logger = logging.getLogger(__name__)



def delegate_action(func):
    @wraps(func)
    def _wrapper(self, input, start, tokens):
        if self.delegate is not None and hasattr(self.delegate, func.func_name):
            try:
                return getattr(self.delegate, func.func_name)(input, start, tokens)
            except:
                raise Exception(repr(sys.exc_info()[1]) + repr(tokens))

    return _wrapper


class RewriteParser(CSSParser):


    def __init__(self, delegate=None):
        super(RewriteParser, self).__init__()
        self.delegate = delegate
        self.import_.setParseAction(self.import_parse_action)
        self.import_location.setParseAction(self.import_location_parse_action)
        self.declaration.setParseAction(self.declaration_parse_action)
        self.declaration_value.setParseAction(self.declaration_value_parse_action)
        # using add here because we want to keep the original
        self.charset_declaration.addParseAction(self.charset_declaration_action)


    @delegate_action
    def import_parse_action(self, input, start, tokens):
        pass


    @delegate_action
    def import_location_parse_action(self, input, start, tokens):
        pass


    @delegate_action
    def declaration_parse_action(self, input, start, tokens):
        pass


    @delegate_action
    def declaration_value_parse_action(self, input, start, tokens):
        pass


    @delegate_action
    def charset_declaration_action(self, input, start, tokens):
        pass



class CSSReference(object):

    def __init__(self, rewriter, filename):
        self.filename = filename
        self.dir = filename.dirname()
        self.rewriter = rewriter


    def exists(self):
        return self.filename.exists()


    def basename(self):
        return self.filename.basename()


    def copy(self, other):
        return self.filename.copy(other)


    def splitext(self):
        return self.filename.splitext()


    def md5(self):
        return self.filename.md5()


    def __hash__(self):
        return hash(str(self.filename))


    def __eq__(self, other):
        return str(self.filename) == str(other.filename)


    def __lt__(self, other):
        return str(self.filename) < str(other.filename)


    def __repr__(self):
        return repr(self.filename)


    def __str__(self):
        return str(self.filename)



class CSSImage(CSSReference):


    def __init__(self, rewriter, filename):
        super(CSSImage, self).__init__(rewriter, filename)
        # the CSS that link to this image
        self.referrers = set()


    def add_referrer(self, css):
        self.referrers.add(css)


    def rewritten_url(self):
        return self.rewriter.rewritten_url(self)




class CSSFile(CSSReference):


    def __init__(self, rewriter, filename, dependencies=()):
        super(CSSFile, self).__init__(rewriter, filename)
        self.unique_name = None

        if self.filename.exists():
            # get the encoding & read the files contents
            # based on that
            charset = "utf-8"
            content = self.filename.open().read()
            if "@charset" in content:
                parser = CSSParser()
                # try & parse a possible charset_declaration
                parser.charset_declaration.parseString(content[:content.index(";")+1], False)
                charset = parser.charset
            self.content = content.decode(charset)
            self.replacer = Replacer(self.content)
            self._replaced_content = None
        self.import_locations = []
        # other CSS referred from this CSS
        self.references = set(dependencies)


    def get_output(self, unix_newlines=False):
        if self._replaced_content is None:
            self._replaced_content = self.replacer.get_value()
            if unix_newlines:
                self._replaced_content = self._replaced_content.replace("\r\n", "\n")
        return self._replaced_content


    def process(self):
        parser = RewriteParser(delegate=self)
        parser.parseString(self.content, True)


    def charset_declaration_action(self, input, loc, tokens):
        # remove the charset declaration, we have utf-8 only
        start, end = tokens[0], tokens[-1]
        self.replacer[start:end] = ""


    def import_parse_action(self, input, start, tokens):
        start, end = tokens[0], tokens[-1]
        tokens = tokens[1:-1]
        position = slice(start, end)
        location = self.import_locations.pop()

        if isinstance(location, StringNode):
            # this *must* be a relative reference,
            # so resolve it
            self.references.add(self.rewriter.add_css(self.filename.dirname() / location.content))
            # and we can safely remove that location
            self.replacer[position] = ""
        elif isinstance(location, URINode):
            if location.relative:
                self.references.add(self.rewriter.add_css(self.filename.dirname() / location.path))
                self.replacer[position] = ""


    def import_location_parse_action(self, input, start, tokens):
        self.import_locations.append(tokens[0])


    def declaration_parse_action(self, input, loc, tokens):
        property_ = tokens[0]
        if property_ in ("background", "background-image"):
            uris = [uri for uri in tokens if isinstance(uri, URINode)]
            if not uris:
                return
            assert len(uris) == 1
            image_url = uris[0]
            if image_url.relative:
                img = self.rewriter.add_image(self.dir / image_url.path)
                img.add_referrer(self)
                if img.exists():
                    self.replacer[image_url.position] = img.rewritten_url


    @classmethod
    def disambiguate_names(cls, css_files):
        """
        Creates a sorting-criteria that will not depend on
        the *full* filenames, but instead only so much of the
        paths as is needed to disambiguate them.
        """
        name_mapping = defaultdict(list)
        working_list = list(css_files)
        level = 1

        def get_name(css_file):
            parts = []
            for _ in xrange(level):
                css_file, right = css_file.split()
                parts.append(right)
            return "/".join(reversed(parts))
        while working_list:
            for css_file in working_list:
                name = get_name(css_file.filename)
                name_mapping[name].append(css_file)

            working_list[:] = []

            for key, value in name_mapping.iteritems():
                if len(value) == 1:
                    value[0].unique_name = key
                else:
                    working_list.extend(value)
            level += 1
            name_mapping.clear()

        for css_file in css_files:
            assert css_file.unique_name is not None, css_file


class SimpleCopyImagePreprocessor(object):


    def process(self, images, basedir, images_path):
        """
        Gets a list of images, and returns a mapping
        of them to their urls.

        Only overwrites images in the destination which
        already exist if they happen to have a differing
        md5-sum.
        """
        res = {}
        for img in images:
            img_md5 = img.md5()
            dest = images_path / img.basename()
            if dest.exists() and dest.md5() != img_md5:
                prefix, ext = img.splitext()
                count = 1
                while True:
                    basename = "%s-%i%s" % (prefix, count, ext)
                    dest = (images_path / basename)
                    if dest.exists() and dest.md5() == img_md5:
                        break
                    if not dest.exists():
                        break
                    count += 1
            if not dest.exists():
                img.copy(dest)
            res[img] = str(dest)[len(str(basedir))+1:] # plus the path-sep
        return res


class MD5ImagePreprocessor(object):


    def process(self, images, basedir, images_path):
        res = {}
        for img in images:
            _, ext = img.splitext()
            dest = images_path / ("%s%s" % (img.md5(), ext))
            if not dest.exists():
                img.copy(dest)
            res[img] = str(dest)[len(str(basedir))+1:] # plus the path-sep
        return res



class CSSRewriter(object):


    def __init__(self,
                 output,
                 images_path="images",
                 image_processor=None,
                 unix_newlines=True,
                 ):
        if image_processor is None:
            image_processor = SimpleCopyImagePreprocessor()
        self.image_processor = image_processor

        self.unix_newlines = unix_newlines
        self.output = output
        self.images_path = self.output.dirname() / images_path
        if self.images_path.exists():
            assert self.images_path.isdir()
        else:
            self.images_path.mkdir()
        self.all_css = set()
        self.css_to_process = []
        self.non_existing = set()
        self.images = {}
        self.image_paths = {}


    def add_css(self, css_file, dependencies=()):
        def process_css(css):
            if css not in self.all_css:
                self.all_css.add(css)
                if not css.exists():
                    self.non_existing.add(css)
                else:
                    self.css_to_process.append(css)
                self.work()
            # we need to return the exact same instance
            return [c for c in self.all_css if css == c][0]

        deps = set()
        for dep in dependencies:
            dep = CSSFile(self, dep)
            dep = process_css(dep)
            deps.add(dep)

        css = CSSFile(self, css_file, deps)
        return process_css(css)


    def add_image(self, imagefile):
        img = CSSImage(self, imagefile)
        if img not in self.images:
            self.images[img] = img
        return self.images[img]


    def work(self):
        while self.css_to_process:
            next_css, self.css_to_process = self.css_to_process[0], self.css_to_process[1:]
            next_css.process()


    def topological_order(self):
        # this is needed so that the sort is stable over
        # various machines
        CSSFile.disambiguate_names(self.all_css)

        def by_unique_name(cf):
            return cf.unique_name

        all_ = sorted(set(css for css in self.all_css if css not in self.non_existing), key=by_unique_name)
        L = []
        S = all_ #set(all_)
        # build up the inverse relationship of edges,
        # as our algorithm works better with these,
        for css in all_:
            css.incoming = []

        for css in all_:
            for adjacence in css.references:
                if adjacence in all_ and css not in adjacence.incoming:
                    adjacence.incoming.append(css)

        S = sorted(set(css for css in all_ if not css.incoming), key=by_unique_name)
        assert S, "Cyclic graph of CSS!"
        while S:
            S.sort()
            n = S.pop()
            L.append(n)
            for m in n.references:
                # no non-existing css-files
                if m in all_:
                    m.incoming.remove(n)
                    if not m.incoming and m not in S:
                        S.append(m)

        if set(css for css in all_ if css.incoming):
            raise Exception("At least one cycle")
        return L


    def rewritten_url(self, img):
        return "url(%s)" % self.image_path_mapping[img]


    def rewrite(self):
        # first, pre-process and copy the images
        # into their destination to get their
        # final relative url.

        basedir = self.output.directory()

        images = []
        for img in self.images.values():
            if not img.exists():
                logger.warn("Image '%r' couldn't be found!", img)
            else:
                images.append(img)
        self.image_path_mapping = self.image_processor.process(images,
                                                               basedir,
                                                               self.images_path
                                                               )

        for css in self.non_existing:
            logger.warn("CSS file '%r' couldn't be found!", css)


        with self.output.open("w") as outf:
            L = reversed(self.topological_order())
            for css in L:
                outf.write("\n/* ---- %s ---- */\n" % css.unique_name)
                outf.write(css.get_output(self.unix_newlines).encode("utf-8"))






"""
Copyright (c) 2009 Ableton AG

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""
import urlparse

from pyparsing import *


# modeled after the grammar found at
# http://www.w3.org/TR/CSS21/grammar.html



class Node(object):

    def __init__(self, position, tokens):
        self.position = position
        self.tokens = tokens



class StringNode(Node):

    def __init__(self, position, tokens):
        super(StringNode, self).__init__(position, tokens)
        self.text = "".join(tokens)
        self.quotechar = self.text[0]
        self.content = self.text[1:-1]


    @classmethod
    def parse_action(cls, input, loc, tokens):
        start, text, end = tokens
        position = slice(start, end)
        return cls(position, [text])


class URINode(Node):

    def __init__(self, position, tokens):
        super(URINode, self).__init__(position, tokens)
        path = tokens[1]
        if isinstance(path, StringNode):
            self.relative = True
            self.path = path.content
        else:
            # we got a string, which is a URL-terminal
            self.path = path
            protocol, _, path, _, _, _ = urlparse.urlparse(path)
            self.relative = protocol is not None and not path.startswith("/")


    @classmethod
    def parse_action(cls, input, loc, tokens):
        position = slice(tokens[0], tokens[-1])
        return cls(position, tokens[1:-1])



class CSSParser(object):


    def __init__(self, robust=False):
        self.charset = "utf-8"

        escapes = r"\\\\|\\\."
        
        MINUS = Literal("-")
        IDENT = Regex(ur"(-|\*)?([a-zA-Z_-]|(%(escapes)s|[\u0200-\u0377]))([a-zA-Z0-9_-]|(%(escapes)s|[\u0200-\u0377]))*" %
                       dict(escapes=escapes))
        COMMA = Literal(",")
        LBRACE = Literal("{")
        RBRACE = Literal("}")
        LSQUARE = Literal("[")
        RSQUARE = Literal("]")
        LPAREN = Literal("(")
        RPAREN = Literal(")")
        IMPORT_SYM = Keyword("@import")
        CHARSET_SYM = Keyword("@charset")
        MEDIA_SYM = Keyword("@media")
        SEMI = Literal(";")
        HASH = Combine(Literal("#") + Regex("([a-zA-Z0-9_-]|(%(escapes)s))+" %
                                            dict(escapes=escapes)))
        ASTERISK = Literal("*")
        DOT = Literal(".")
        COLON = Literal(":")
        PLUS = Literal("+")
        GREATER = Literal(">")
        SLASH = Literal("/")
        IMPORTANT_SYM = Literal("!") + Keyword("important") # TODO: comment missing
        EQUAL = Literal("=")

        STRING = originalTextFor(quotedString)
        STRING.setParseAction(StringNode.parse_action)

        INCLUDES = Literal("~=")
        DASHMATCH = Literal("|=")
        #FUNCTION = Combine(IDENT+ LPAREN)
        URL = OneOrMore(Word(srange("[!#$%&*-~]"))) # TODO

        URI = originalTextFor(Literal("url(") + (STRING | URL) + Literal(")"))
        URI.setParseAction(URINode.parse_action)

        NUMBER = Regex(r"[0-9]*\.[0-9]+|[0-9]+")
        #NUMBER = Combine(OneOrMore(numlit)) ^ Combine(ZeroOrMore(numlit) + DOT + OneOrMore(numlit))

        PERCENTAGE = Combine(NUMBER + Literal("%"))
        # LENGTH = Combine(NUMBER + (Literal("px") ^ Literal("cm") ^ Literal("mm") ^
        #                                Literal("in") ^ Literal("pt") ^ Literal("pc")))
        LENGTH = Combine(NUMBER + (Literal("px") | Literal("cm") | Literal("mm") |
                                       Literal("in") | Literal("pt") | Literal("pc")))
        EMS = Combine(NUMBER + Literal("em"))
        EXS =Combine(NUMBER + Literal("ex"))
        ANGLE = Combine(NUMBER + (Literal("deg") | Literal("rad") | Literal("grad")))
        TIME = Combine(NUMBER + (Literal("ms") | Literal("s")))
        FREQ = Combine(NUMBER + (Literal("hz") | Literal("khz")))

        # BNF
        medium = IDENT

        import_location = STRING | URI

        import_ = originalTextFor(IMPORT_SYM + import_location + Optional(medium + ZeroOrMore(COMMA + medium)) + SEMI)

        unary_operator = MINUS | PLUS

        operator =  SLASH | COMMA

        hexcolor = HASH

        expr = Forward()

        # this is not according to the
        #function = FUNCTION + ZeroOrMore(Optional(IDENT + EQUAL) + expr) + RPAREN
        funcarg = Optional(IDENT + EQUAL) + expr
        function = IDENT + LPAREN + Group(Optional(delimitedList(funcarg))) + RPAREN

        term = Optional(unary_operator) + ((PERCENTAGE | LENGTH | EMS | EXS | ANGLE | TIME | FREQ | NUMBER)\
                                           | STRING | URI | hexcolor | (function ^ IDENT))

        expr << (term + ZeroOrMore( Optional(operator) +  term))

        property_ = IDENT

        prio = IMPORTANT_SYM

        declaration_value = expr + Optional(prio)
        if robust:
            declaration_value = Optional(declaration_value)

        declaration = property_ + COLON + declaration_value

        element_name = IDENT | ASTERISK

        class_ = Combine(DOT + IDENT)

        attrib = LSQUARE + IDENT + Optional((EQUAL ^ INCLUDES ^ DASHMATCH ) + ( IDENT ^ STRING ) ) + RSQUARE

        pseudo = COLON + ( IDENT ^ (IDENT + LPAREN + Optional(funcarg) + RPAREN))

        simple_selector = (element_name + (ZeroOrMore( HASH | class_ | attrib | pseudo )) |
                           OneOrMore( HASH | class_ | attrib | pseudo))

        combinator = PLUS | GREATER

        selector = simple_selector + ZeroOrMore(Optional(combinator) + simple_selector)

        ruleset = selector + ZeroOrMore(COMMA + selector) + \
                  LBRACE + Optional(declaration) + ZeroOrMore(SEMI + Optional(declaration)) + RBRACE

        media = MEDIA_SYM + medium + ZeroOrMore(COMMA + medium) + LBRACE + ruleset + RBRACE

        charset_declaration = originalTextFor(CHARSET_SYM + STRING + SEMI)
        charset_declaration.setParseAction(self.set_charset)

        # TODO-dir: implement the "page"-stuff
        content = media | ruleset
        # The SEMI below is just because some stylesheets are "dirty"
        if robust:
            content =  content ^ SEMI
        stylesheet = Optional(charset_declaration) +\
                     ZeroOrMore(import_) +\
                     ZeroOrMore(content) +\
                     StringEnd()


        stylesheet.ignore(cppStyleComment)
        stylesheet.ignore(htmlComment)

        for name in ("stylesheet", "import_", "URI", "import_location", "declaration",
                     "declaration_value", "NUMBER", "function", "expr", "LENGTH", "ANGLE",
                     "TIME", "FREQ", "term", "selector", "charset_declaration", "STRING",
                     "IDENT",
                     ):
            setattr(self, name, locals()[name])



    def set_charset(self, input, loc, tokens):
        self.charset = tokens[2].content.lower()


    def parseString(self, string, parseAll=False):
        # we need this because otherwise replacement
        # won't work - it works with the "raw" bytes,
        # not with expanded tabs.
        self.stylesheet.parseWithTabs()
        return self.stylesheet.parseString(string, parseAll)




if __name__ == "__main__":
    import sys
    parser = CSSParser()
    if len(sys.argv) == 2:
        parser.parseString(open(sys.argv[1]).read(), True)
        sys.exit(0)


    print parser.stylesheet.parseString("""@import url(../foo);""", parseAll=True)



import os
import sys
from setuptools import setup, find_packages

setup(
    name="abl.cssprocessor",
    version="0.2",
    description="A processor to aggregate CSS-files and their referenced media",
    author="Diez B. Roggisch",
    author_email="diez.roggisch@ableton.com",
    url="",
    license="MIT",
    download_url='',
    install_requires=[
        "pyparsing",
        "abl.vpath>=0.4",
        "argparse",
        ],
    packages=find_packages(exclude=['ez_setup', 'tests']),
    namespace_packages = ['abl', ],
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Environment :: Web Environment :: ToscaWidgets',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Widget Sets',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    entry_points= {

        "console_scripts": [
            "css_validator = abl.cssprocessor.validator:validator",
            ]
        }
)



"""
Copyright (c) 2009 Ableton AG

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""
from __future__ import absolute_import, with_statement

import sys
import pprint
from textwrap import dedent

import argparse
from pyparsing import ParseException

from abl.vpath.base import URI

from .parser import URINode
from .rewriter import RewriteParser




class CSSValidator(object):


    def __init__(self, filename, verbose=False, roots=()):
        self.filename = filename

        root_mapping = []
        for path in roots:
            if isinstance(path, tuple):
                # for testing, you can pass the mapping as (prefix, URI)
                root_mapping.append(path)
            else:
                if ":" in path:
                    prefix, path = path.split(":", 1)
                else:
                    prefix = "/"
                assert path.startswith("/"), "%s must start with /"
                root_mapping.append((prefix, URI("file://%s" % path)))

        self.root_mapping = root_mapping
        self.verbose = verbose
        self.warnings = []


    def validate(self):
        if not self.filename.exists() or self.filename.isdir():
            self.error("No such file: %r" % self.filename)

        self.relative_base = self.filename.directory()
        self.report("Attempting to parse CSS file %s" % self.filename)
        parser = RewriteParser(self)
        with self.filename.open() as inf:
            content = inf.read()
        try:
            res = parser.parseString(content)
        except ParseException, e:
            message = "ERROR: Malformed CSS, text is: <<<%s>>>" % e.pstr
            self.warnings.append(message)
            print message



    def declaration_parse_action(self, input, loc, tokens):
        property_ = tokens[0]
        if property_ in ("background", "background-image"):
            uris = [uri for uri in tokens if isinstance(uri, URINode)]
            if not uris:
                return
            assert len(uris) == 1
            image_url = uris[0]
            path = image_url.path
            if image_url.relative:
                if not (self.relative_base / path).exists():
                    self.warning("Relative resource not found: %s" % path)
            else:
                if not self.check_for_absoulte_resource(path):
                    self.warning("Absolute resource not found: %s" % path)


    def check_for_absoulte_resource(self, resource_path):
        for prefix, path in self.root_mapping:
            if resource_path.startswith(prefix):
                rest = resource_path[len(prefix):]
                if rest.startswith("/"):
                    rest = rest[1:]
                if (path / rest).exists():
                    return True
        return False


    def error(self, msg):
        sys.stderr.write(msg)
        sys.stderr.write("\n")
        sys.exit(1)


    def report(self, msg):
        if self.verbose:
            print msg


    def warning(self, msg):
        self.warnings.append(msg)
        print "WARNING:", msg


def validator():
    parser = argparse.ArgumentParser()
    parser.add_argument("cssfile",
                        help=dedent("""
                        The CSS-file to validate.
                        """),
                        )

    parser.add_argument("--root",
                        help=dedent("""
                        To resolve absolute paths, you need to give
                        at least on root-path.

                        You can specify several.

                        It's also possible to give a prefix by using
                        a colon.

                        So

                          --root=/resources:/home/user/project/public/resources

                        will be used to resolve resources that start with '/resources'.
                        """),
                        action="append",
                        dest="roots",
                        default=[],
                       )
    parser.add_argument("-v", "--verbose", action="store_true",
                        default=False,)



    args = parser.parse_args(sys.argv[1:])


    validator = CSSValidator(URI("file://%s" % args.cssfile),
                             roots=args.roots,
                             verbose=args.verbose,
                             )

    validator.validate()

